
// implements RFC862 server using /dev/tunX
//

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <linux/if_tun.h>

#include "flags.h"

unsigned char Ip_frame[IP_FRAME_MAX_SIZE+1];

static int Done=0; // end prog and/or signal number
void tschues(int goodbyval){
    Done=goodbyval;
}

int reply_to_echo_req(int fd,void *frame,int len){
   int i;
   u16_t chksum;
   u32_t u32;
   u8_t *data;
   int ip_hl; // in octets
   unsigned char ttl;
   data=(u8_t*)frame;
   if (len<A_IP_HEADER_OCTET_MIN_LENGTH+A_ICMP_HEADER_ECHO_OCTET_LEN)
       return -1; // if it is to short it is no echo request
   if (a_ip_get_version(data)!=4)
       return -4;
   ip_hl=4*a_ip_get_header_length(data);
   if (ip_hl<A_IP_HEADER_OCTET_MIN_LENGTH)
       return -2; // not minmal lentgh
   if (swap16(*a16(data,A_IP_OFFSET_LENGTH))!=len)
       return -3;
   if (*a8(data,ip_hl+A_ICMP_OFFSET_TYPE )!=0x8)
       return -5; // no echo request
   ttl=*a8(data,A_IP_OFFSET_TTL);
   if (!ttl)
       return -6;
   // modify icmp header
   *a8(data,ip_hl+A_ICMP_OFFSET_TYPE)=0; // echo reply
   *a16(data,ip_hl+A_ICMP_OFFSET_CHKSUM)=0; //chksum=0
   chksum=ip4_chk_sum(data+ip_hl,len-ip_hl);
   *a16(frame,ip_hl+A_ICMP_OFFSET_CHKSUM)=~chksum;
   //
   // modify ip header
   *a8(data,A_IP_OFFSET_TTL)=--ttl;
   // exchange src and destination
   u32=*a32(data,A_IP_OFFSET_SRC_ADDR);
   *a32(data,A_IP_OFFSET_SRC_ADDR)=*a32(data,A_IP_OFFSET_DEST_ADDR);
   *a32(data,A_IP_OFFSET_DEST_ADDR)=u32;
   //
   *a16(data,A_IP_OFFSET_CHKSUM)=0; //chksum=0
   chksum=ip4_chk_sum(data,ip_hl);
   *a16(data,A_IP_OFFSET_CHKSUM)=~chksum;
   //
   //print_ip_frame("out_frame",frame,len);
   i=write(fd,frame,len);
   return 0; //ok
};

int main(int argc, char **argv){
    int fd;
    fd_set rfds;
    fd_set efds;
    struct timeval tv;
    int i;
    int n;
    char *dev_new_name;
    if (argc!=5){
        fprintf(stderr,"usage:\n%s /dev/net/tun local_ipv4_addr remote_net remote_netmask\n",argv[0]);
        exit(1);
    }
    fd=tun_alloc(argv[1],&dev_new_name);
    if (fd<0){
        fprintf(stderr,"\nfatal: open(%s,O_RDWR) failed\n",argv[1]);
        exit(1);
    }
    ioctl(fd,TUNSETNOCSUM,IFF_TUN); // stuff to discuss
    //
    // now assign the ipv4 address
//fprintf(stderr,"\n%d",__LINE__);
    snprintf(Ip_frame,sizeof(Ip_frame),"ifconfig %s inet %s mtu 576",dev_new_name,argv[2]);
    fprintf(stdout,"\nnow executing:\n%s",Ip_frame);
    system(Ip_frame);
    fprintf(stdout,"\nifconfig done\n");
    //
    //
    snprintf(Ip_frame,sizeof(Ip_frame),
        "route add -net %s netmask %s dev %s",argv[3],argv[4],dev_new_name);
    fprintf(stdout,"\nnow executing:\n%s",Ip_frame);
    system(Ip_frame);
    fprintf(stdout,"\nroute done\n");
    n=1+max(2,fd);
    for (;!Done;){
        FD_ZERO(&rfds);
        FD_SET(fd,&rfds);
        FD_ZERO(&efds);
        FD_SET(fd,&efds);
        tv.tv_sec=0;
        tv.tv_usec=10000;
        select(n,&rfds,NULL,&efds,&tv);
        if (FD_ISSET(fd,&rfds)){
            i=read(fd,Ip_frame,sizeof(Ip_frame));
            if (i>0){
                //print_ip_frame("in_frame",Ip_frame,i);
                // see if is an echo req
                i=reply_to_echo_req(fd,Ip_frame,i);
                if (i<0)
                    fprintf(stderr,"\nerror frame is not echo, reason=%3d",i);
            }
        }
    }
    return Done;
}


