
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <linux/if_tun.h>

#include <curses.h>
#include "flags.h"

// just ab bit of version info
#define P_VERSION "V0.01"
#define COMPILE_TIME    __TIME__
#define COMPILE_DATE    __DATE__

// restore terminal before asserting
#define ASSERT(a) do { endwin(); assert(a); } while (0)


// waste a bit of (virtual) memory for coding convenience
// I have 1 GByte so less than 900 kByte should not hurt anyway
static unsigned char Blocks[FL_BLOCKS_PER_IP_FRAME_MAX][FL_BLOCK_MAX_SIZE];
static unsigned char Ip_frame[IP_FRAME_MAX_SIZE];

static int Done=0; // end prog and/or signal number
void tschues(int goodbyval){
    Done=goodbyval;
}



// that is not the usual curses way
void emit_string(int yy, int xx, const char *s){
    int c;
    int x;
    int y;
    x=xx; y=yy;
    ASSERT((s!=NULL) && (y>=0) && (x>=0));
    move(y,x);
    while ((c=*s++)!='\0'){
        if (c=='\n'){
            y++;x=xx;
            move(y,x);
        }else if ((c==' ') || isprint(c)){
            addch(c);
        }else
            ASSERT(1);
    }
    getyx(stdscr,y,x);
    refresh();
}

void block_print(int yy, int xx,unsigned char a[FL_BLOCK_MAX_SIZE],int max_size){
    int i;
    int c;
    int x;
    int y;
    x=xx; y=yy;
    ASSERT((a!=NULL) && (y>=0) && (x>=0));
    move(y,x);
    for (i=0;i<max_size;){
        c=a[i];
        if (c<0x16)
            c+='a';
        addch(c);
        i++;
        if (i%32==0){
            y++;x=xx;
            move(y,x);
        }else if (i%16==0){
            addch(' ');addch(' ');addch(' ');//addch(' ');addch(' ');
        }else if (i%8==10000){
            addch(' ');addch(' ');addch(' ');
        }else if (i%4==0){
            addch(' ');
        }
        //refresh();
        if ((c==FL_BL_FIN) || (i>=max_size))
            return;
    }
    getyx(stdscr,y,x);
    refresh();
}

// brief help text
static const char *User_info=
    "ip-flags " P_VERSION " compiled " COMPILE_DATE " " COMPILE_TIME "\n"
    "usage: flagsin [tun-device]\n"
    "    tun-device ... defaults to /dev/net/tunX\n\n"
    "keyboard input commands:\n"
    "   1 ... undo (current) nipple\n"
    "   2 ... undo (current) block\n"
    "   3 ... undo (current) ipframe\n"
    "   4 ... commit block (can NOT be undone!)\n"
    "   5 ... commit and send ip-frame\n"
    "   [abcdefghijklmnop] ... ordinary flags for nipples\n"
    "   s ... start block\n"
    "   t ... end block\n"
    "   Q ... quit program\n";

// also del trailing space
void yx_print(int yy, int xx,const char *fmt,...){
    char sbuff[100];
    int n;
    int i;
    int c;
    int x,y;
    va_list ap;
    va_start(ap,fmt);
    n = vsnprintf (sbuff,100, fmt, ap);
    getyx(stdscr,y,x);
    move(yy,xx);
    for (i=0;i<n;i++){
        c=sbuff[i];
        if (c==0)
            c=' ';
        addch(c);
    }
    move(y,x);
    va_end(ap);
}

char Ht[]="abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"t";

#define POS_X_HELP  1
#define POS_Y_HELP  1
#define POS_X_ERR   1
#define POS_Y_ERR   15
#define POS_X_BLOCK 1
#define POS_Y_BLOCK 17
#define POS_X_INFO  60
#define POS_Y_INFO  1

int valid_block(int i){
    return i>i-1; //always tre checks are missing here
}

int main(int argc, char **argv){
    int fd;
    int i;
    int j;
    char *dev_new_name;
    int c;
    int x;
    int y;
    int blk_num;
    int flag_num;
    int frame_num;
    if (argc!=5){
        fprintf(stderr,"usage:\n%s /dev/net/tun local_ipv4_addr remote_net remote_netmask\n",argv[0]);
        exit(1);
    }
    fd=tun_alloc(argv[1],&dev_new_name);
    if (fd<0){
        fprintf(stderr,"\nfatal: open(%s,O_RDWR) failed\n",argv[1]);
        exit(1);
    }
    ioctl(fd,TUNSETNOCSUM,IFF_TUN); // stuff to discuss
    snprintf(Ip_frame,sizeof(Ip_frame),"ifconfig %s inet %s mtu 576",dev_new_name,argv[2]);
    fprintf(stdout,"\nnow executing:\n%s",Ip_frame);
    system(Ip_frame);
    fprintf(stdout,"\nifconfig done\n");

    snprintf(Ip_frame,sizeof(Ip_frame),
        "route add -net %s netmask %s dev %s",argv[3],argv[4],dev_new_name);
    fprintf(stdout,"\nnow executing:\n%s",Ip_frame);
    system(Ip_frame);
    fprintf(stdout,"\nroute done\n");
    //
    fcntl(0,F_SETFL,O_NDELAY);
    initscr();
    signal(SIGINT,tschues);
    noecho();
    cbreak();
    keypad(stdscr,TRUE);
    x=y=0;
    emit_string(POS_Y_HELP,POS_X_HELP,User_info);
    getyx(stdscr,y,x);
    //ipv4_2_fl(Ht,sizeof Ht,200,Blocks);
    //block_print(12, 0,&Blocks[0][0]);
    //block_print(18, 0,Ht);
    frame_num=0;
    blk_num=0;
    flag_num=0;
    while (!Done){
        c=getch();
        if (c==EOF){
            usleep(10000L);
            errno=0;
            continue;
        }
        clear();
        emit_string(POS_Y_HELP,POS_X_HELP,User_info);
        yx_print(POS_Y_INFO  ,POS_X_INFO,"frame_num=%4d",frame_num);
        yx_print(POS_Y_INFO+1,POS_X_INFO,"blk_num  =%4d",blk_num);
        yx_print(POS_Y_INFO+2,POS_X_INFO,"flag_num =%4d",flag_num);
        switch(c){
	case 'H':
	    {
	        char* h_string = "aaaaaefaaaaboaacpeaaaeaabcgjdakaaaaboakaaaaaaaiaaoglibbbgaadaaaab";
		char* cp = h_string;
		while (*cp)
		    Blocks[blk_num][flag_num++] = *cp++-'a';
	    }
	    break;
        case '1':
            // undo nipple
            if (flag_num>0){
                flag_num--;
            }else{
                yx_print(POS_Y_ERR,POS_X_ERR,"nothing to undo");
            }
            break;
        case '2':
            // undo block
            if (flag_num==0 && blk_num==0){
                yx_print(POS_Y_ERR,POS_X_ERR,"nothing to undo");
            }else if (flag_num>0){
                flag_num=0;
                if (blk_num>0)
                    blk_num--;
            }
            break;
        case '3':
            // undo ipframe
            if ((flag_num<=0) && (blk_num<=0)){
                yx_print(POS_Y_ERR,POS_X_ERR,"nothing to undo");
            }else{
                flag_num=0;
                blk_num=0;
            }
            break;
        case '4':
            // commit block
            if (blk_num<0){
                yx_print(POS_Y_ERR,POS_X_ERR,"no block to commit");
            }else if (Blocks[blk_num][0]!=FL_BL_START){
                yx_print(POS_Y_ERR,POS_X_ERR,"block does not start with start with FL_BL_START");
            }else if (Blocks[blk_num][flag_num-1]!=FL_BL_FIN){
                yx_print(POS_Y_ERR,POS_X_ERR,"block does not end with FL_BL_FIN");
            }else if (blk_num<FL_BLOCKS_PER_IP_FRAME_MAX){
                blk_num++;
                flag_num=0;
            }else{
                yx_print(POS_Y_ERR,POS_X_ERR,"error: commiting block %s:%d",__FILE__,__LINE__);
            }
            break;
        case '5':
            // commit and send ip-frame
            if ((blk_num<=0) || !valid_block(blk_num-1)){
                yx_print(POS_Y_ERR,POS_X_ERR,"noting (valid) to commit");
                break;
            }
            i=fl_2_ipv4(Ip_frame,blk_num,Blocks);
            if (i<20){
                yx_print(POS_Y_ERR,POS_X_ERR,"error: sizeof(ipframe)<20 ? - nothing sent");
                break;
            }
            j=write(fd,Ip_frame,i);
            if (i!=j){
                yx_print(POS_Y_ERR,POS_X_ERR,"error: sendig ipframe: %d=write(...,%d",j,i);
            }else{
                yx_print(POS_Y_ERR,POS_X_ERR,"info: %d octets sent for ipframe",j);
                blk_num=0;
                flag_num=0;
                frame_num++;
            }
            break;
        case 'a': case 'b': case 'c': case 'd':
        case 'e': case 'f': case 'g': case 'h':
        case 'i': case 'j': case 'k': case 'l':
        case 'm': case 'n': case 'o': case 'p':
            if (flag_num==0){
               yx_print(POS_Y_ERR,POS_X_ERR,"error: start must be at the begin");
            }else if (flag_num>=FL_BLOCK_MAX_SIZE-2){
                yx_print(POS_Y_ERR,POS_X_ERR,"error: to many flags in block");
            }else if (Blocks[blk_num][flag_num-1]==FL_BL_FIN){
                yx_print(POS_Y_ERR,POS_X_ERR,"error: there may be no data after block end");
            }else{
                Blocks[blk_num][flag_num++]=c-'a';
            }
            break;
        case 's':
            // start block
            if (blk_num>=FL_BLOCKS_PER_IP_FRAME_MAX){
                yx_print(POS_Y_ERR,POS_X_ERR,"error: to many blocks");
            }else if (flag_num!=0){
                yx_print(POS_Y_ERR,POS_X_ERR,"error: block already started");
            }else{
                Blocks[blk_num][flag_num++]=FL_BL_START;
            }
            break;
        case 't':
            // end block
            if (flag_num==0){
               yx_print(POS_Y_ERR,POS_X_ERR,"error: end can not be at the begin");
            }else if (Blocks[blk_num][flag_num-1]==FL_BL_FIN){
                yx_print(POS_Y_ERR,POS_X_ERR,"error: there may be no block end after block end");
            }else if (blk_num>=FL_BLOCKS_PER_IP_FRAME_MAX){
                yx_print(POS_Y_ERR,POS_X_ERR,"error: to many blocks in ip-frame");
            }else if (flag_num>=FL_BLOCK_MAX_SIZE-1){
                yx_print(POS_Y_ERR,POS_X_ERR,"error: to many flags in block");
            }else{
                Blocks[blk_num][flag_num++]=FL_BL_FIN;
            }
            break;
        case 'Q':
            Done=1;
            break;
        default:
            if (isprint(c) && !isspace(c))
                yx_print(POS_Y_ERR,POS_X_ERR,"error: invalid key pressed 0x%02x %c",c,c);
            else
                yx_print(POS_Y_ERR,POS_X_ERR,"error: invalid key pressed 0x%02x",c);
        }
        //move(y,x);
        //getyx(stdscr,y,x);
        yx_print(POS_Y_INFO  ,POS_X_INFO,"frame_num=%4d",frame_num);
        yx_print(POS_Y_INFO+1,POS_X_INFO,"blk_num  =%4d",blk_num);
        yx_print(POS_Y_INFO+2,POS_X_INFO,"flag_num =%4d",flag_num);
        block_print(POS_Y_BLOCK,POS_X_BLOCK,&Blocks[blk_num][0],flag_num);
        move(y,x);
        refresh();
    }
    endwin();
    return Done;
}

