
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <linux/if_tun.h>

#include <curses.h>
#include "flags.h"

// just ab bit of version info
#define P_VERSION "V0.01"
#define COMPILE_TIME    __TIME__
#define COMPILE_DATE    __DATE__

// restore terminal before asserting
#define ASSERT(a) do { endwin(); assert(a); } while (0)


// waste a bit of (virtual) memory for coding convenience
// I have 1 GByte so less than 900 kByte should not hurt anyway
static unsigned char Blocks[FL_BLOCKS_PER_IP_FRAME_MAX][FL_BLOCK_MAX_SIZE];
static unsigned char Ip_frame[IP_FRAME_MAX_SIZE];

static int Done=0; // end prog and/or signal number
void tschues(int goodbyval){
    Done=goodbyval;
}


struct S_CHAIN{
    struct S_CHAIN *next;
    struct S_CHAIN *prev;
    void *data;
    int   dlen; // data length
};

int Ip_frames_qued=0;
struct S_CHAIN Ip_tail;
struct S_CHAIN Ip_head={&Ip_tail,NULL,NULL,0};
struct S_CHAIN Ip_tail={NULL,&Ip_head,NULL,0};

struct S_CHAIN *ip_frame_que(void *data,int len){
     struct S_CHAIN *header;
     header=malloc(sizeof (struct S_CHAIN));
     if (!header)
         return NULL;
     header->data=malloc(len);
     if (!header->data){
         free(header);
         return NULL;
     }
     header->dlen=len;
     //
     header->prev=&Ip_head;
     header->next=Ip_head.next;
     Ip_head.next=header;
     header->next->prev=header;
     //
     memcpy(header->data,data,len);
     Ip_frames_qued++;
     return header;
}

// you must separately free the data
void *ip_frame_unque(int *len){
    void *data;
    struct S_CHAIN *header;
    if (Ip_frames_qued<=0)
        return NULL;
    header=Ip_tail.prev;
    data=header->data;
    *len=header->dlen;
    //
    Ip_tail.prev=header->prev;
    header->prev->next=&Ip_tail;
    //
    Ip_frames_qued--;
    free(header);
    return data;
}

// that is not the usual curses way
void emit_string(int yy, int xx, const char *s){
    int c;
    int x;
    int y;
    x=xx; y=yy;
    ASSERT((s!=NULL) && (y>=0) && (x>=0));
    move(y,x);
    while ((c=*s++)!='\0'){
        if (c=='\n'){
            y++;x=xx;
            move(y,x);
        }else if ((c==' ') || isprint(c)){
            addch(c);
        }else
            ASSERT(1);
    }
    getyx(stdscr,y,x);
    refresh();
}

void block_print(int yy, int xx,unsigned char a[FL_BLOCK_MAX_SIZE],int max_size){
    int i;
    int c;
    int x;
    int y;
    x=xx; y=yy;
    ASSERT((a!=NULL) && (y>=0) && (x>=0));
    move(y,x);
    for (i=0;i<max_size;){
        c=a[i];
        if (c<0x16)
            c+='a';
        addch(c);
        i++;
        if (i%32==0){
            y++;x=xx;
            move(y,x);
        }else if (i%16==0){
            addch(' ');addch(' ');addch(' ');//addch(' ');addch(' ');
        }else if (i%8==0){
            addch(' ');addch(' ');addch(' ');
        }else if (i%4==0){
            addch(' ');
        }
        //refresh();
        if ((c==FL_BL_FIN) || (i>=max_size))
            break;
    }
    getyx(stdscr,y,x);
    refresh();
}

// brief help text
static const char *User_info=
    "ip-flags " P_VERSION " compiled " COMPILE_DATE " " COMPILE_TIME "\n"
    "usage: flagsout [tun-device]\n"
    "    tun-device ... defaults to /dev/net/tunX\n\n"
    "keyboard input commands:\n"
    "   p ... previous block\n"
    "   n ... next block\n"
    "   i ... next ipframe (there is no way to go back to the previous ip-frame!\n"
    "   Q ... quit program\n";

// also del trailing space
void yx_print(int yy, int xx,const char *fmt,...){
    char sbuff[100];
    int n;
    int i;
    int c;
    int x,y;
    va_list ap;
    va_start(ap,fmt);
    n = vsnprintf (sbuff,100, fmt, ap);
    getyx(stdscr,y,x);
    move(yy,xx);
    for (i=0;i<n;i++){
        c=sbuff[i];
        if (c==0)
            c=' ';
        addch(c);
    }
    move(y,x);
    va_end(ap);
}

char Ht[]="abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"t";

#define POS_X_HELP  1
#define POS_Y_HELP  1
#define POS_X_ERR   1
#define POS_Y_ERR   15
#define POS_X_BLOCK 1
#define POS_Y_BLOCK 17
#define POS_X_INFO  60
#define POS_Y_INFO  1

int valid_block(int i){
    return i>i-1; //always the checks are missing here
}

int main(int argc, char **argv){
    int fd;
    fd_set rfds;
    fd_set efds;
    unsigned char *ipfp=NULL;
    struct timeval tv;
    int n;
    int l;
    int k;
    char *dev_new_name;
    int c;
    int x;
    int y;
    int blk_num;
    int flag_num;
    int frame_num;
    int bc; // block count for present ip frame;
    if (argc!=5){
        fprintf(stderr,"usage:\n%s /dev/net/tun local_ipv4_addr remote_net remote_netmask\n",argv[0]);
        exit(1);
    }
    fd=tun_alloc(argv[1],&dev_new_name);
    if (fd<0){
        fprintf(stderr,"\nfatal: open(%s,O_RDWR) failed\n",argv[1]);
        exit(1);
    }
    ioctl(fd,TUNSETNOCSUM,IFF_TUN); // stuff to discuss
    snprintf(Ip_frame,sizeof(Ip_frame),"ifconfig %s inet %s mtu 576",dev_new_name,argv[2]);
    fprintf(stdout,"\nnow executing:\n%s",Ip_frame);
    system(Ip_frame);
    fprintf(stdout,"\nifconfig done\n");

    snprintf(Ip_frame,sizeof(Ip_frame),
        "route add -net %s netmask %s dev %s",argv[3],argv[4],dev_new_name);
    fprintf(stdout,"\nnow executing:\n%s",Ip_frame);
    system(Ip_frame);
    fprintf(stdout,"\nroute done\n");
    //
    fcntl(0,F_SETFL,O_NDELAY);
    initscr();
    signal(SIGINT,tschues);
    noecho();
    cbreak();
    keypad(stdscr,TRUE);
    x=y=0;
    emit_string(POS_Y_HELP,POS_X_HELP,User_info);
    getyx(stdscr,y,x);
    frame_num=0;
    blk_num=0;
    flag_num=0;
    bc=0;
    n=1+max(2,fd);
    k=0;
    while (!Done){
        k++;
        // get ip frame
        FD_ZERO(&rfds);
        FD_SET(fd,&rfds);
        FD_ZERO(&efds);
        FD_SET(fd,&efds);
        tv.tv_sec=0;
        tv.tv_usec=10000;
        select(n,&rfds,NULL,&efds,&tv);
        if (FD_ISSET(fd,&rfds)){
	    //yx_print(POS_Y_ERR+1,POS_X_ERR,"S");
            //read in a new ip-frame
            l=read(fd,Ip_frame,IP_FRAME_MAX_SIZE);
            if (l>0){
	        //yx_print(POS_Y_ERR+1,POS_X_ERR,"q                 ");
                ip_frame_que(Ip_frame,l);
                //yx_print(POS_Y_ERR+1,POS_X_ERR+1,"()");
		//refresh();
            }else{
                // error
            }
        }
        // get input character
        c=getch();
        if (c==EOF){
            //usleep(10000L); no sleep, required due to previous select
            errno=0;
            continue;
        }
        clear();
        emit_string(POS_Y_HELP,POS_X_HELP,User_info);
        yx_print(POS_Y_INFO  ,POS_X_INFO,"frame_num     =%8d",frame_num);
        yx_print(POS_Y_INFO+1,POS_X_INFO,"blk_num       =%8d",blk_num);
        yx_print(POS_Y_INFO+2,POS_X_INFO,"flag_num      =%8d",flag_num);
        yx_print(POS_Y_INFO+3,POS_X_INFO,"qued_ip_frames=%8d",Ip_frames_qued);
        switch(c){
        case 'p':
            // undo block
            if (blk_num>0){
                blk_num--;
            }else{
                yx_print(POS_Y_ERR,POS_X_ERR,"nothing to undo");
            }
            break;
        case 'n':
            // next block
            if (blk_num<bc-1){
                blk_num++;
            }else{
                yx_print(POS_Y_ERR,POS_X_ERR,"no next block in ip frame");
            }
            break;
        case 'i':
            // next ipframe
            //yx_print(POS_Y_ERR+1,POS_X_ERR+4,"u");
	    if (ipfp)
	        free(ipfp);
            ipfp=ip_frame_unque(&l);
            //yx_print(POS_Y_ERR+1,POS_X_ERR+5,"()");
            if (ipfp && l){
                // convert ip frame to blocks
                bc=ipv4_2_fl(ipfp,l,120/*flags_per_junk*/,Blocks);
                if (bc<=0){
                    // error
                    bc=0;
                }else{
                    blk_num=0;
                    frame_num++;
                }

            }else{
                bc=0;
                blk_num=0;
            }
            break;
        case 'Q':
            Done=1;
            break;
        default:
            if (isprint(c) && !isspace(c))
                yx_print(POS_Y_ERR,POS_X_ERR,"error: invalid key pressed 0x%02x %c",c,c);
            else
                yx_print(POS_Y_ERR,POS_X_ERR,"error: invalid key pressed 0x%02x",c);
        }
        //move(y,x);
        //getyx(stdscr,y,x);
        yx_print(POS_Y_INFO  ,POS_X_INFO,"frame_num     =%8d",frame_num);
        yx_print(POS_Y_INFO+1,POS_X_INFO,"blk_num       =%8d",blk_num);
        yx_print(POS_Y_INFO+2,POS_X_INFO,"flag_num      =%8d",flag_num);
        yx_print(POS_Y_INFO+3,POS_X_INFO,"qued_ip_frames=%8d",Ip_frames_qued);
        yx_print(POS_Y_INFO+4,POS_X_INFO,"k             =%8d",k);
        block_print(POS_Y_BLOCK,POS_X_BLOCK,&Blocks[blk_num][0],FL_BLOCK_MAX_SIZE);
        move(y,x);
        refresh();
    }
    endwin();
    return Done;
}


