
//we put common stuff here, but only if it will not use curses

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <linux/if_tun.h>

#include "flags.h"

int tun_alloc(char *dev,char **new_name){
    struct ifreq ifr;
    int fd;
    int err;
    if (!dev || !new_name)
        return -1;
    *new_name=NULL;
    if ((fd=open(dev,O_RDWR))<0)
        return fd;
    memset(&ifr,0,sizeof(ifr));
    ifr.ifr_flags = IFF_TUN|IFF_NO_PI;
    if ((err=ioctl(fd,TUNSETIFF,(void*)&ifr))<0){
        close(fd);
        return err;
    }
    *new_name=strdup(ifr.ifr_name);
    return fd;
}

int print_ip_frame(char *text,void *frame,int len){
    unsigned char *d;
    int n;
    int c;
    d=(unsigned char*)frame;
    c=fprintf(stdout,"\n%s %d octets",text,len);
    for (n=0;n<len;n++){
        if (!(n&3))
            c+=fprintf(stdout,"\n");
        if (n==4*5)
            c+=fprintf(stdout,"\n");
        c+=fprintf(stdout," %02x",d[n]);
        if (n==len-1)
            c+=fprintf(stdout,"\n");
    }
    return c;
}

// calculate header check sum
unsigned short ip4_chk_sum(void *base, int len){
    int sum;
    int i;
    unsigned char *p;
    p=(unsigned char*)base;
    sum=0;
    for (i=0;i<len;i++){
        if (i&1)
            sum+=*p++ &0xff;
        else
            sum+=(*p++ &0xff)*256;
    }
    sum&=0xffff;
    return swap16(sum);
}

int ipv4_2_fl(void *ip_frame, int ip_frame_size,int flags_per_junk,
    unsigned char a[FL_BLOCKS_PER_IP_FRAME_MAX][FL_BLOCK_MAX_SIZE]){
    int ic_ip; // in count counting from ip_frame
    int ic_bl; // in count counting from fl_block
    int bc; // block_count
    int nc; // nipple count in individual block
    int i;
    if (ip_frame_size<20)
        return 0;
    if (ip_frame_size>IP_FRAME_MAX_SIZE)
        return 0;
    if (flags_per_junk<20)
        flags_per_junk=20;
    if (flags_per_junk>FL_BLOCK_PAYLOAD_MAX_SIZE)
        flags_per_junk=FL_BLOCK_PAYLOAD_MAX_SIZE;
    bc=0;
    ic_ip=0;
    while (ic_ip<2*ip_frame_size){
        if (bc>=FL_BLOCKS_PER_IP_FRAME_MAX)
            return 0;
        nc=0;
        // start
        a[bc][FL_START_OFFSET]=FL_BL_START;
        // type
        a[bc][FL_BLOCK_TYPE_OFFSET]=FL_DATA;
        // protocol
        a[bc][FL_PROT_OFFSET]=FL_PROTOCOL_IP4;
        // checksum
        a[bc][FL_CHECKSUM_TYPE_OFFSET]=FL_CHECKSUM_NONE;
        // blocknum - we correct it later
        a[bc][FL_BLOCKNUM_OFFSET_HIGH]=nipple_hi(bc);
        a[bc][FL_BLOCKNUM_OFFSET_LOW]=nipple_lo(bc);
        ic_bl=0;
        nc=FL_DATA_OFFSET;
        while(ic_bl<flags_per_junk && ic_ip<2*ip_frame_size){
            a[bc][nc++]=get_nipple(ip_frame,ic_ip);
            ic_bl++;
            ic_ip++;
        }
        //a[bc][nc++]=FL_BL_DATA_END;
        //
        // we use no checksum
        //
        a[bc][nc++]=FL_BL_FIN;
        bc++;
    }
    a[bc][nc++]=FL_IP_FRAME_FIN;
	// block numbers are counting down - so fix it
    for (i=0;i<bc;i++){
        a[i][FL_BLOCKNUM_OFFSET_HIGH]=nipple_hi(bc-i-1);
        a[i][FL_BLOCKNUM_OFFSET_LOW ]=nipple_lo(bc-i-1);
    }
    return bc; // number of blocks
}

int fl_get_frame_len(unsigned char *blk){
	int i;
	for (i=0;*blk++!=FL_BL_FIN && i<FL_BLOCK_MAX_SIZE ;i++)
		;
	return i;
}

int fl_2_ipv4(void *ip_frame, int blocks,
    unsigned char a[FL_BLOCKS_PER_IP_FRAME_MAX][FL_BLOCK_MAX_SIZE]){
    int bl;
    int nc;
    int c;
    int ic_ip;
    unsigned char *ipf;
    ipf=(unsigned char*)ip_frame;
    if (blocks<=0 || blocks>256)
        return 0;
    ic_ip=0;
    for (bl=0;bl<blocks;bl++){
         // most block checks are done elsewhere
         for (nc=FL_DATA_OFFSET;nc<FL_BLOCK_MAX_SIZE;nc++){
              c=a[bl][nc];
              if (c==FL_BL_FIN)
                  break;
              if (c&0xf0){
                  return 0; // maleformed block (at least at this pint in time)
              }
              if (nc&1){
                  ipf[ic_ip]|=c;
                  ic_ip++;
              }else{
                  ipf[ic_ip]=c<<4;
              }
         }
    }
    return ic_ip; // ipv4 length
}

int nipple_hi(int i){
    return (i>>4)&0xf;
}

int nipple_lo(int i){
    return i&0xf;
}

int get_nipple(void *base,int index){
    unsigned char *c;
    c=(unsigned char*)base;
    if (index&1)
        return  c[index/2]     &0xf;
    else
        return (c[index/2]>>4) &0xf;
}
