
//
// flagio.c
//
// Combination of Aaron's flagin.c and flagout.c
//

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <linux/if_tun.h>

#include <curses.h>
#include "flags.h"

// just ab bit of version info
#define P_VERSION "V0.11"
#define COMPILE_TIME    __TIME__
#define COMPILE_DATE    __DATE__
#define TRANSMITTING 1

// restore terminal before asserting
#define ASSERT(a) do { endwin(); assert(a); } while (0)

static int Done=0; // end prog and/or signal number
void tschues(int goodbyval){
    Done=goodbyval;
}



// that is not the usual curses way
void emit_string(int yy, int xx, const char *s){
    int c;
    int x;
    int y;
    x=xx; y=yy;
    ASSERT((s!=NULL) && (y>=0) && (x>=0));
    move(y,x);
    while ((c=*s++)!='\0'){
        if (c=='\n'){
            y++;x=xx;
            move(y,x);
        }else if ((c==' ') || isprint(c)){
            addch(c);
        }else
            ASSERT(1);
    }
    getyx(stdscr,y,x);
    refresh();
}

void block_print(int yy, int xx,unsigned char a[FL_BLOCK_MAX_SIZE],int max_size){
    int i;
    int c;
    int x;
    int y;
    x=xx; y=yy;
    ASSERT((a!=NULL) && (y>=0) && (x>=0));
    move(y,x);
    for (i=0;i<max_size;){
        c=a[i];
        if (c<0x16)
            c+='a';
        addch(c);
        i++;
        if (i%32==0){
            y++;x=xx;
            move(y,x);
        }else if (i%16==0){
            addch(' ');addch(' ');addch(' ');//addch(' ');addch(' ');
        }else if (i%8==10000){
            addch(' ');addch(' ');addch(' ');
        }else if (i%4==0){
            addch(' ');
        }
        //refresh();
        if ((c==FL_BL_FIN) || (i>=max_size))
            return;
    }
    getyx(stdscr,y,x);
    refresh();
}

// brief help text
static const char *Receiving_info=
	"### R E C E I V I N G ###\n"
    "flagio " P_VERSION " compiled " COMPILE_DATE " " COMPILE_TIME "\n"
    "usage: flagsin [tun-device]\n"
    "    tun-device ... defaults to /dev/net/tunX\n\n"
    "keyboard input commands:\n"
    "   s ... undo (current) nipple\n"
    "   t ... undo (current) block\n"
    "   ? ... undo (current) ipframe\n"
    "   1 ... commit block (can NOT be undone!)\n"
    "   9 ... commit and send ip-frame\n"
    "   [abcdefghijklmnop] ... ordinary flags for nipples\n"
    "   q ... start block\n"
    "   r ... end block\n"
    "   z ... end ip-frame\n"
	"   # ... Toggle transmitting/receiving\n"
    "   $ ... quit program\n";

// also del trailing space
void yx_print(int yy, int xx,const char *fmt,...){
    char sbuff[100];
    int n;
    int i;
    int c;
    int x,y;
    va_list ap;
    va_start(ap,fmt);
    n = vsnprintf (sbuff,100, fmt, ap);
    getyx(stdscr,y,x);
    move(yy,xx);
    for (i=0;i<n;i++){
        c=sbuff[i];
        if (c==0)
            c=' ';
        addch(c);
    }
    move(y,x);
    va_end(ap);
}

char Ht[]="abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"abc223bbbabcbab3b34bbabb3bb4bb32b2b2bbbvseddderrssaserddasehb"
"t";

#define POS_X_HELP  1
#define POS_Y_HELP  1
#define POS_X_ERR   1
#define POS_Y_ERR   20
#define POS_X_BLOCK 1
#define POS_Y_BLOCK 22
#define POS_X_INFO  60
#define POS_Y_INFO  1

int valid_block(int i){
    return i>i-1; //almost always true checks are missing here ;)
}

struct S_CHAIN{
    struct S_CHAIN *next;
    struct S_CHAIN *prev;
    void *data;
    int   dlen; // data length
};

int Ip_frames_qued=0;
struct S_CHAIN Ip_tail;
struct S_CHAIN Ip_head={&Ip_tail,NULL,NULL,0};
struct S_CHAIN Ip_tail={NULL,&Ip_head,NULL,0};

struct S_CHAIN *ip_frame_que(void *data,int len){
     struct S_CHAIN *header;
     header=malloc(sizeof (struct S_CHAIN));
     if (!header)
         return NULL;
     header->data=malloc(len);
     if (!header->data){
         free(header);
         return NULL;
     }
     header->dlen=len;
     //
     header->prev=&Ip_head;
     header->next=Ip_head.next;
     Ip_head.next=header;
     header->next->prev=header;
     //
     memcpy(header->data,data,len);
     Ip_frames_qued++;
     return header;
}

// you must separately free the data
void *ip_frame_unque(int *len){
    void *data;
    struct S_CHAIN *header;
    if (Ip_frames_qued<=0)
        return NULL;
    header=Ip_tail.prev;
    data=header->data;
    *len=header->dlen;
    //
    Ip_tail.prev=header->prev;
    header->prev->next=&Ip_tail;
    //
    Ip_frames_qued--;
    free(header);
    return data;
}

// brief help text
static const char *Transmitting_info=
	"### T R A N S M I T T I N G ###\n"
    "flagio " P_VERSION " compiled " COMPILE_DATE " " COMPILE_TIME "\n"
    "usage: flagio tun-device local_ipv4_addr remote_ip4v_addr\n"
    "    tun-device ... defaults to /dev/net/tunX\n\n"
    "keyboard input commands:\n"
    "   1 ... previous block\n"
    "   9 ... next block\n"
    "   5 ... next ipframe (there is no way to go back to the previous ip-frame!\n"
	"   # ... Toggle transmitting/receiving\n"
    "   $ ... quit program\n";

static struct {
	// transmit
 	unsigned char *ipfp;
    unsigned char blocks[FL_BLOCKS_PER_IP_FRAME_MAX][FL_BLOCK_MAX_SIZE];
	unsigned char ip_frame[IP_FRAME_MAX_SIZE];
    int blk_num;
    int flag_num;
    int frame_num;
    int l;
    int k;
    int bc; // block count for present ip frame;
	int i;
	int j;
	int flags_per_chunk;
}T={NULL};

static struct {
	// receive
    unsigned char blocks[FL_BLOCKS_PER_IP_FRAME_MAX][FL_BLOCK_MAX_SIZE];
	unsigned char ip_frame[IP_FRAME_MAX_SIZE];
    int blk_num;
    int flag_num;
    int frame_num;
    int l;
    int k;
    int bc; // block count for present ip frame;
	int i;
	int j;
}R;

static struct {
	int k;
}Co={0};

int transmit(int c) {
	int x;
	int y;
	emit_string(POS_Y_HELP,POS_X_HELP,Transmitting_info);
	switch(c){
	case 't':
		// undo block
		if (T.blk_num>0){
			T.blk_num--;
		}else{
			yx_print(POS_Y_ERR,POS_X_ERR,"nothing to undo");
		}
		break;
	case '9':
		// next block
		if (T.blk_num<T.bc-1){
			T.blk_num++;
		}else{
			yx_print(POS_Y_ERR,POS_X_ERR,"no next block in ip frame");
		}
		break;
	case '5':
		// next ipframe
		//yx_print(POS_Y_ERR+1,POS_X_ERR+4,"u");
		if (T.ipfp)
			free(T.ipfp);
		T.ipfp=ip_frame_unque(&T.l);
		//yx_print(POS_Y_ERR+1,POS_X_ERR+5,"()");
		if (T.ipfp && T.l){
			// convert ip frame to blocks
			T.bc=ipv4_2_fl(T.ipfp,T.l,T.flags_per_chunk,T.blocks);
			if (T.bc<=0){
				// error
				T.bc=0;
			}else{
				T.blk_num=0;
				T.frame_num++;
			}
		}else{
			T.bc=0;
			T.blk_num=0;
		}
		break;
	case '$':
		Done=1;
		break;
	case EOF:
		break;
	default:
		if (isprint(c) && !isspace(c))
			yx_print(POS_Y_ERR,POS_X_ERR,"error: invalid key pressed 0x%02x %c",c,c);
		else
			yx_print(POS_Y_ERR,POS_X_ERR,"error: invalid key pressed 0x%02x",c);
	}
	emit_string(POS_Y_HELP,POS_X_HELP,Transmitting_info);
	yx_print(POS_Y_INFO  ,POS_X_INFO,"frame_num      =%10d",T.frame_num);
	yx_print(POS_Y_INFO+1,POS_X_INFO,"blk_num        =%10d",T.blk_num);
	yx_print(POS_Y_INFO+2,POS_X_INFO,"flag_num       =%10d",T.flag_num);
	yx_print(POS_Y_INFO+3,POS_X_INFO,"qued_ip_frames =%10d",Ip_frames_qued);
	yx_print(POS_Y_INFO+4,POS_X_INFO,"rem_blk_in_ip  =%10d",T.bc-T.blk_num-1);
	yx_print(POS_Y_INFO+5,POS_X_INFO,"k              =%10d",Co.k);
	if (T.bc) {
		block_print(POS_Y_BLOCK,POS_X_BLOCK,&T.blocks[T.blk_num][0],FL_BLOCK_MAX_SIZE);
	}
	if (T.bc-T.blk_num-1==0){
    	getyx(stdscr,y,x);
		yx_print(y, x, "z");
	}
	return TRANSMITTING;
}

int receive(int fd,int c) {
	emit_string(POS_Y_HELP,POS_X_HELP,Receiving_info);
	switch(c){
	case 'H':
		{
		char* h_string = "aaaaaefaaaaboaacpeaaaeaabcgjdakaaaaboakaaaaaaaiaaoglibbbgaadaaaab";
		char* cp = h_string;
		while (*cp)
			R.blocks[R.blk_num][R.flag_num++] = *cp++-'a';
		}
		break;
	case 's':
		// undo nipple
		if (R.flag_num>0){
		R.flag_num--;
		}else{
		    yx_print(POS_Y_ERR,POS_X_ERR,"nothing to undo");
		}
		break;
	case 't':
		// undo block
		if (R.flag_num==0 && R.blk_num==0){
		    yx_print(POS_Y_ERR,POS_X_ERR,"nothing to undo");
		}else if (R.flag_num>0){
		    R.flag_num=0;
		    if (R.blk_num>0)
		 	   R.blk_num--;
		}
		break;
	case '?':
		// undo ipframe
		if ((R.flag_num<=0) && (R.blk_num<=0)){
		    yx_print(POS_Y_ERR,POS_X_ERR,"nothing to undo");
		}else{
		    R.flag_num=0;
		    R.blk_num=0;
		}
		break;
	case '1':
		// commit block
		if (R.blk_num<0){
		    yx_print(POS_Y_ERR,POS_X_ERR,"no block to commit");
		}else if (R.blocks[R.blk_num][0]!=FL_BL_START){
		    yx_print(POS_Y_ERR,POS_X_ERR,"block does not start with start with FL_BL_START");
		}else if (R.blocks[R.blk_num][R.flag_num-1]!=FL_BL_FIN){
		    yx_print(POS_Y_ERR,POS_X_ERR,"block does not end with FL_BL_FIN");
		}else if (R.blk_num<FL_BLOCKS_PER_IP_FRAME_MAX){
		    R.blk_num++;
		    R.flag_num=0;
		}else{
		    yx_print(POS_Y_ERR,POS_X_ERR,"error: commiting block %s:%d",__FILE__,__LINE__);
		}
		break;
	case '9':
		// commit and send ip-frame
		if ((R.blk_num<=0) || !valid_block(R.blk_num-1)){
		    yx_print(POS_Y_ERR,POS_X_ERR,"noting (valid) to commit");
		    break;
		}
		R.i=fl_2_ipv4(R.ip_frame,R.blk_num,R.blocks);
		if (R.i<20){
		    yx_print(POS_Y_ERR,POS_X_ERR,"error: sizeof(ipframe)<20 ? - nothing sent");
		    break;
		}
		R.j=write(fd,R.ip_frame,R.i);
		if (R.i!=R.j){
		    yx_print(POS_Y_ERR,POS_X_ERR,"error: sendig ipframe: %d=write(...,%d",R.j,R.i);
		}else{
		    yx_print(POS_Y_ERR,POS_X_ERR,"info: %d octets sent for ipframe",R.j);
		    R.blk_num=0;
		    R.flag_num=0;
		    R.frame_num++;
		}
		break;
	case 'a': case 'b': case 'c': case 'd':
	case 'e': case 'f': case 'g': case 'h':
	case 'i': case 'j': case 'k': case 'l':
	case 'm': case 'n': case 'o': case 'p':
		if (R.flag_num==0){
		   yx_print(POS_Y_ERR,POS_X_ERR,"error: start must be at the begin");
		}else if (R.flag_num>=FL_BLOCK_MAX_SIZE-2){
		    yx_print(POS_Y_ERR,POS_X_ERR,"error: to many flags in block");
		}else if (R.blocks[R.blk_num][R.flag_num-1]==FL_BL_FIN){
		    yx_print(POS_Y_ERR,POS_X_ERR,"error: there may be no data after block end");
		}else{
		    R.blocks[R.blk_num][R.flag_num++]=c-'a';
		}
		break;
	case 'q':
		// start block
		if (R.blk_num>=FL_BLOCKS_PER_IP_FRAME_MAX){
		    yx_print(POS_Y_ERR,POS_X_ERR,"error: to many blocks");
		}else if (R.flag_num!=0){
		    yx_print(POS_Y_ERR,POS_X_ERR,"error: block already started");
		}else{
		    R.blocks[R.blk_num][R.flag_num++]=FL_BL_START;
		}
		break;
	case 'r':
		// end block
		if (R.flag_num==0){
		   yx_print(POS_Y_ERR,POS_X_ERR,"error: end can not be at the begin");
		}else if (R.blocks[R.blk_num][R.flag_num-1]==FL_BL_FIN){
		    yx_print(POS_Y_ERR,POS_X_ERR,"error: there may be no block end after block end");
		}else if (R.blk_num>=FL_BLOCKS_PER_IP_FRAME_MAX){
		    yx_print(POS_Y_ERR,POS_X_ERR,"error: to many blocks in ip-frame");
		}else if (R.flag_num>=FL_BLOCK_MAX_SIZE-1){
		    yx_print(POS_Y_ERR,POS_X_ERR,"error: to many flags in block");
		}else{
		    R.blocks[R.blk_num][R.flag_num++]=FL_BL_FIN;
			yx_print(POS_Y_ERR,POS_X_ERR,"use '1 to commit block");
		}
		break;
	case '$':
		Done=1;
		break;
	case EOF:
		break;
	case 'z':
		   yx_print(POS_Y_ERR,POS_X_ERR,"use '9 to commit frame");
		break;
	default:
		if (isprint(c) && !isspace(c))
		   yx_print(POS_Y_ERR,POS_X_ERR,"error: invalid key pressed 0x%02x %c",c,c);
		else
		   yx_print(POS_Y_ERR,POS_X_ERR,"error: invalid key pressed 0x%02x",c);
	}
	emit_string(POS_Y_HELP,POS_X_HELP,Receiving_info);
	yx_print(POS_Y_INFO  ,POS_X_INFO,"frame_num     =%8d",R.frame_num);
	yx_print(POS_Y_INFO+1,POS_X_INFO,"blk_num       =%8d",R.blk_num);
	yx_print(POS_Y_INFO+2,POS_X_INFO,"flag_num      =%8d",R.flag_num);
	block_print(POS_Y_BLOCK,POS_X_BLOCK,&R.blocks[R.blk_num][0],R.flag_num);
	return !TRANSMITTING;
}

int main(int argc, char **argv){
    char *dev_new_name;
	char buf[1024];
    struct timeval tv;
    int fd;
    fd_set rfds;
    fd_set efds;
	int transmitting;
	int x;
	int y;
	int n;
	int c;

    if ((argc!=4)&&(argc!=5)){
        fprintf(stderr,"usage:\n%s /dev/net/tun local_ipv4_addr remote_ipv4_addr [flags_per_chunk between 60 and 300]\n",argv[0]);
        exit(1);
    }
	if (argc==5){
		T.flags_per_chunk=atoi(argv[4]);
		if ((T.flags_per_chunk < 60)||(T.flags_per_chunk > 300)) {
	        fprintf(stderr,"%s: invalid number for flags_per_chunk\n", argv[4]);
			exit(1);
		}
	} else {
		T.flags_per_chunk=120;
	}
    fd=tun_alloc(argv[1],&dev_new_name);
    if (fd<0){
        fprintf(stderr,"\nfatal: open(%s,O_RDWR) failed\n",argv[1]);
        exit(1);
    }
    ioctl(fd,TUNSETNOCSUM,IFF_TUN); // stuff to discuss
    //snprintf(Ip_frame,sizeof(Ip_frame),"ifconfig %s inet %s mtu 576",dev_new_name,argv[2]);
    snprintf(buf,sizeof(buf),"ifconfig %s %s pointopoint %s mtu 576",dev_new_name,argv[2],argv[3]);
    fprintf(stdout,"\nnow executing:\n%s",buf);
    system(buf);
    fprintf(stdout,"\nifconfig done\n");
    fcntl(0,F_SETFL,O_NDELAY);
    initscr();
	cbreak();
	noecho();
	nonl();
    intrflush(stdscr, FALSE);
    keypad(stdscr, TRUE);
    signal(SIGINT,tschues);
    x=y=0;
    emit_string(POS_Y_HELP,POS_X_HELP,Transmitting_info);
    getyx(stdscr,y,x);
    n=1+max(2,fd);
    //k=0;
	transmitting=!TRANSMITTING;
	clear();
	transmitting=receive(fd,EOF);	
    while (!Done){
        Co.k++;
        // get ip frame
        FD_ZERO(&rfds);
        FD_SET(fd,&rfds);
        FD_ZERO(&efds);
        FD_SET(fd,&efds);
        tv.tv_sec=0;
        tv.tv_usec=10000;
        select(n,&rfds,NULL,&efds,&tv);
        if (FD_ISSET(fd,&rfds)){
            T.l=read(fd,T.ip_frame,IP_FRAME_MAX_SIZE);
            if (T.l>0){
                ip_frame_que(T.ip_frame,T.l);
            }else{
                // error
            }
        }
     // get input character
        c=getch();
        if (c==EOF){
            //usleep(10000L); no sleep, required due to previous select
            errno=0;
            continue;
        }else if(c=='#'){
			transmitting=!transmitting;
			c=EOF;
		}
        clear();
		if (transmitting) {
			transmit(c);
		} else {
			receive(fd,c);
		}
		move(y,x);
	    refresh();
    }
    endwin();
    return Done;
}

